import sys
import json


class ServoMove(object):

	_LEGAL_TYPE = ("action", "loop", "posture")
	_LEGAL_POSITION_RANGE = (0, 180)
	_LEGAL_SPEED = ("high", "med", "low")
	_SPEED_TO_STEP = {
		'high': 180,
		'med': 18,
		'low': 6
	}

	def __init__(self, type=None, startPosition=None, endPosition=None, speed="med", totalTimes=float('INF')):
		# configuration variables
		self.type = type
		self.startPosition = startPosition
		self.endPosition = endPosition
		self.speed = speed
		self.totalTimes = totalTimes
		self.verify()
		# performing variables
		self.performedTimes = 0
		self.currentPosition = None
		self.stepVector = None
		self.goingForward = None

	def verify(self):
		# type
		if self.type not in ServoMove._LEGAL_TYPE:
			raise ValueError('move type should be one of the following: %s (%s given).' % (str(ServoMove._LEGAL_TYPE), self.type))
		# startPosition
		if not isinstance(self.startPosition, int):
			raise TypeError('startPosition should be an integer. (%s given)' % (str(self.startPosition)))
		if not (ServoMove._LEGAL_POSITION_RANGE[0] <= self.startPosition <= ServoMove._LEGAL_POSITION_RANGE[1]):
			raise ValueError('startPosition should be inside range: %s, and %d given' % (ServoMove._LEGAL_POSITION_RANGE, self.startPosition))
		# endPosition
		if self.type != "posture":
			if not isinstance(self.endPosition, int):
				raise TypeError('endPosition should be an integer. (%s given)' % (str(self.endPosition)))
			elif not (ServoMove._LEGAL_POSITION_RANGE[0] <= self.endPosition <= ServoMove._LEGAL_POSITION_RANGE[1]):
				raise ValueError('endPosition should be inside range: %s, and %d given' % (ServoMove._LEGAL_POSITION_RANGE, self.endPosition))
		# speed
		if self.speed not in ServoMove._LEGAL_SPEED:
			raise ValueError('move speed should be one of the following: %s (%s given).' % (str(ServoMove._LEGAL_SPEED), self.speed))

	def getNext(self):
		"""
			This function should be called 10 times / sec.
		"""
		if self.currentPosition is None:
			self.initCurrentPosition()
		else:
			self.updateCurrentPosition()
		return self.currentPosition
	
	def initCurrentPosition(self):
		self.currentPosition = self.startPosition
		if self.type == 'action' or self.type == 'loop':
			self.stepVector = ServoMove._SPEED_TO_STEP[self.speed] * (self.endPosition - self.startPosition) / abs(self.endPosition - self.startPosition)
			self.goingForward = True
		elif self.type == 'posture':
			self.stepVector = 0
		else:
			raise ValueError('cannot recognize this type: ' + self.type)

	def updateCurrentPosition(self):
		if self.type == 'action' or self.type == 'loop':
			if self.type == 'action' and self.performedTimes > 0:
				pass
			elif self.type == 'loop' and self.performedTimes > self.totalTimes:
				pass
			elif self.goingForward:
				if abs(self.stepVector) >= abs(self.endPosition - self.currentPosition):
					self.currentPosition = self.endPosition
					self.goingForward = False
					self.performedTimes += 1
				else:
					self.currentPosition += self.stepVector
			else:
				if abs(self.stepVector) >= abs(self.startPosition - self.currentPosition):
					self.currentPosition = self.startPosition
					self.goingForward = True
					self.performedTimes += 1
				else:
					self.currentPosition -= self.stepVector
		elif self.type == 'posture':
			pass
		else:
			raise ValueError('cannot recognize this type: ' + self.type)

	def isFinished(self):
		ret = (self.type == 'posture')
		ret |= (self.type == 'action' and self.performedTimes > 0)
		ret |= (self.type == 'loop' and self.performedTimes > self.totalTimes)
		return ret

	def reset(self):
		self.performedTimes = 0
		self.currentPosition = None
		self.stepVector = None
		self.goingForward = None

class Move(object):

	# finish criteria (pseudo enum)
	_FINISH_CRITERIA_ANY = 0
	_FINISH_CRITERIA_ALL = 1

	def __init__(self, fromServoMoves=None, fromFile=None, fromServoConfig=None):

		self.servoMoves = []

		if fromServoMoves is not None:
			if not all(isinstance(sm, ServoMove) for sm in fromServoMoves):
				raise ValueError('(Move) __init__(): bad argument (fromServoMoves)')
			self.servoMoves = fromServoMoves
		elif fromFile is not None:
			try:
				fileContent = json.load(open(fromFile, 'r'))
			except IOError as err:
				print('cannot open file: %s (%s)' % (fromFile, err))
				sys.exit(1)
			except ValueError as err:
				print('bad JSON file: %s (%s)' % (fromFile, err))
				sys.exit(1)
			except:
				print('error loading file (unknown)')
				sys.exit(1)
			if not type(fileContent) == type([]):
				raise ValueError('not an array (%s)' % (fromFile))
			self.servoMoves = [ServoMove(**d) for d in fileContent]
		elif fromServoConfig is not None:
			self.servoMoves = [ServoMove(**d) for d in fromServoConfig]
		else:
			raise ValueError('class Move must be instantiate using a file or an array of ServoMove')

	def getNext(self):
		"""
			This function should be called 10 times / sec.
		"""
		return [s.getNext() for s in self.servoMoves]

	def getNumServos(self):
		return len(self.servoMoves)

	def isFinished(self, criteria=_FINISH_CRITERIA_ANY):
		if criteria == Move._FINISH_CRITERIA_ANY:
			return any(s.isFinished() for s in self.servoMoves)
		elif criteria == Move._FINISH_CRITERIA_ALL:
			return all(s.isFinished() for s in self.servoMoves)
		else:
			raise ValueError('cannot recognize finish criteria: %s' % (str(criteria)))

	def reset(self):
		for sm in self.servoMoves:
			sm.reset()



from setuptools import setup

setup(name='roboemotion',
	version='1.0.4',
	description='emotion expressing for robotic arms',
	url='https://github.com/chengjovy/RoboEmotion',
	author='Jovy Cheng',
	author_email='chengjovy83@gmail.com',
	license='MIT',
	packages=['roboemotion'],
	install_requires=[
		'serial',
	],
	package_data={
		'roboemotion': ['roboemotion/robo_emotion_config.json',],
	},
	include_package_data=True,
	zip_safe=False)



from __future__ import print_function
from abc import ABCMeta, abstractmethod
import serial, sys

class Correspondent(object):
	__metaclass__ = ABCMeta

	def __init__(self):
		pass

	@abstractmethod
	def connect(self):
		pass

	@abstractmethod
	def isConnected(self):
		pass

	@abstractmethod
	def write(self, msg='', end=''):
		pass

class SerialCorrespondent(Correspondent):

	def __init__(self, serialIdentifier=None, baudRate=None):
		self.serialIdentifier = serialIdentifier
		self.baudRate = baudRate
		self.device = None

	def connect(self):
		try:
			self.device = serial.Serial(self.serialIdentifier, self.baudRate)
		except Exception as e:
			print('cannot initiate serial connection: ' + str(e))

	def isConnected(self):
		return isinstance(self.device, serial.serialposix.Serial)

	def write(self, msg='', end=''):
		if not self.isConnected():
			raise Exception('not connected')
		self.device.write(msg + end)

class StandardErrorCorrespondent(Correspondent):

	def __init__(self):
		pass

	def connect(self):
		pass

	def isConnected(self):
		return True

	def write(self, msg='', end=''):
		print(msg, end=end, file=sys.stderr)



from __future__ import print_function

import sys, json, thread, pkg_resources
from time import sleep
from util import SerialCorrespondent, StandardErrorCorrespondent
from move import Move


class RoboEmotion(object):

	_DEFAULT_EMOTION_MOVES = {
		"happy":    None,
		"sad":      None,
		"fear":     None,
		"anger":    None,
		"surprise": None,
		"disgust":  None
	}
	_DEFAULT_EMOTION_CONFIG = "robo_emotion_config.json"

	def __init__(self, defaultEmotionConfig=None, correspondent=None):
		#
		# initialize _DEFAULT_EMOTION_MOVES
		#
		if defaultEmotionConfig is not None:
			RoboEmotion._DEFAULT_EMOTION_CONFIG = defaultEmotionConfig
		f = pkg_resources.resource_stream(__name__, RoboEmotion._DEFAULT_EMOTION_CONFIG)
		# fileContent = json.load(open(RoboEmotion._DEFAULT_EMOTION_CONFIG, 'r'))
		fileContent = json.load(f)
		try:
			assert type(fileContent) == type({})
		except:
			raise ValueError("should be a dictionary mapping emotion name to its move (%s)" % (RoboEmotion._DEFAULT_EMOTION_CONFIG))
		for emotionName in fileContent:
			RoboEmotion._DEFAULT_EMOTION_MOVES[emotionName.lower()] = Move(fromServoConfig=fileContent[emotionName])
		#
		# initialize correspondent if necessary
		#
		if correspondent:
			self.correspondent = correspondent
		else:
			self.correspondent = StandardErrorCorrespondent()
		#
		# performing
		#
		self.currentMove = None
		self.finishCriteria = None
		#
		# start communication
		#
		thread.start_new_thread(self._communicate, ())

	def perform(self, move, finishCriteria=Move._FINISH_CRITERIA_ALL, noReset=False):
		if noReset:
			self.currentMove = move
		else:
			move.reset()
			self.currentMove = move
		self.finishCriteria = finishCriteria

	def performEmotion(self, emotionName, finishCriteria=Move._FINISH_CRITERIA_ALL):
		emotionMove = RoboEmotion._DEFAULT_EMOTION_MOVES.get(emotionName.lower())
		if emotionMove is not None:
			emotionMove.reset()
			self.currentMove = emotionMove
			self.finishCriteria = finishCriteria
		else:
			raise ValueError('cannot recognize this emotion: %s' % (emotionName))

	def getNext(self):
		if self.currentMove is not None:
			return self.currentMove.getNext()
		else:
			return RoboEmotion._DEFAULT_EMOTION_MOVES['init'].getNext()

	def isFinished(self):
		return (self.currentMove is None) or (self.currentMove.isFinished(criteria=self.finishCriteria))

	def _communicate(self):
		lastPositions = None
		while True:
			positions = self.getNext()
			positions = " ".join([str(p) for p in positions])
			if lastPositions is None or lastPositions != positions:
				lastPositions = positions
				self.correspondent.write(positions + "\n")
			sleep(0.1)

def communicate(serialCorrespondent=None):
	if roboEmotion is not None:
		lastPositions = None
		while True:
			positions = roboEmotion.getNext()
			positions = " ".join([str(p) for p in positions])
			if lastPositions is None or lastPositions != positions:
				lastPositions = positions
				print(positions)
				if serialCorrespondent is not None:
					serialCorrespondent.write(positions + "\n")
			sleep(0.1)

if __name__ == "__main__":
	import tkinter as tk
	roboEmotion = RoboEmotion()
	try:
		serialCorrespondent = SerialCorrespondent(serialIdentifier=sys.argv[1], baudRate=9600)
	except:
		serialCorrespondent = SerialCorrespondent(serialIdentifier="/dev/cu.usbmodem95", baudRate=9600)
	serialCorrespondent.connect()

	root = tk.Tk()

	w = tk.Button(root, text='init', command=lambda: roboEmotion.performEmotion('init'))
	w.pack(padx=10)
	w = tk.Button(root, text='happy', command=lambda: roboEmotion.performEmotion('happy'))
	w.pack(padx=10)
	w = tk.Button(root, text='sad', command=lambda: roboEmotion.performEmotion('sad'))
	w.pack(padx=10)
	w = tk.Button(root, text='fear', command=lambda: roboEmotion.performEmotion('fear'))
	w.pack(padx=10)
	w = tk.Button(root, text='anger', command=lambda: roboEmotion.performEmotion('anger'))
	w.pack(padx=10)
	w = tk.Button(root, text='surprise', command=lambda: roboEmotion.performEmotion('surprise'))
	w.pack(padx=10)
	w = tk.Button(root, text='disgust', command=lambda: roboEmotion.performEmotion('disgust'))
	w.pack(padx=10)

	thread.start_new_thread(communicate, (serialCorrespondent,))
	# start Tkinter
	root.mainloop()
